import time
from typing import Any, Dict, List, Sequence

import httpx
from opentelemetry.sdk import trace as otel
from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult
from opentelemetry.trace import format_span_id, format_trace_id

from latitude_telemetry.exporter.payloads import (
    Attribute,
    AttributeValue,
    CreateTraceRequestBody,
    Event,
    Link,
    Resource,
    ResourceSpan,
    ScopeSpan,
    Span,
    Status,
)
from latitude_telemetry.telemetry.types import GatewayOptions
from latitude_telemetry.util import Model

RETRIABLE_STATUSES = [408, 409, 429, 500, 502, 503, 504]


class ExporterOptions(Model):
    api_key: str
    gateway: GatewayOptions
    retries: int
    delay: float
    timeout: float


class Exporter(SpanExporter):
    _options: ExporterOptions

    def __init__(self, options: ExporterOptions):
        self._options = options

    def export(self, spans: Sequence[otel.ReadableSpan]) -> SpanExportResult:
        if not spans:
            return SpanExportResult.SUCCESS

        try:
            self._send(self._serialize(list(spans)))

        except Exception:
            return SpanExportResult.FAILURE

        return SpanExportResult.SUCCESS

    def shutdown(self) -> None:
        return None

    def force_flush(self, timeout_millis: int = 30000) -> bool:
        return True

    def _serialize(self, spans: List[otel.ReadableSpan]) -> List[ResourceSpan]:
        if not spans:
            return []

        return [
            ResourceSpan(
                resource=Resource(attributes=self._serialize_attributes(dict(spans[0].resource.attributes or {}))),
                scope_spans=[
                    ScopeSpan(
                        spans=[
                            Span(
                                # Note: span.context should not be None
                                trace_id=format_trace_id(span.context.trace_id if span.context else 0),
                                # Note: span.context should not be None
                                span_id=format_span_id(span.context.span_id if span.context else 0),
                                parent_span_id=format_span_id(span.parent.span_id) if span.parent else None,
                                name=span.name,
                                kind=span.kind.value,
                                # Note: span.start_time should not be None
                                start_time=str(span.start_time or 0),
                                end_time=str(span.end_time) if span.end_time else None,
                                status=Status(
                                    code=span.status.status_code.value,
                                    message=span.status.description,
                                ),
                                events=[
                                    Event(
                                        name=event.name,
                                        time=str(event.timestamp),
                                        attributes=self._serialize_attributes(dict(event.attributes or {})),
                                    )
                                    for event in span.events
                                ],
                                links=[
                                    Link(
                                        trace_id=format_trace_id(link.context.trace_id),
                                        span_id=format_span_id(link.context.span_id),
                                        attributes=self._serialize_attributes(dict(link.attributes or {})),
                                    )
                                    for link in span.links
                                ],
                                attributes=self._serialize_attributes(dict(span.attributes or {})),
                            )
                            for span in spans
                        ]
                    )
                ],
            )
        ]

    def _serialize_attributes(self, attributes: Dict[str, Any]) -> List[Attribute]:
        serialized_attributes: List[Attribute] = []

        for key, value in attributes.items():
            serialized_value = AttributeValue(string=str(value))

            if isinstance(value, str):
                serialized_value = AttributeValue(string=value)

            elif isinstance(value, bool):
                serialized_value = AttributeValue(boolean=value)

            elif isinstance(value, int):
                serialized_value = AttributeValue(integer=value)

            elif isinstance(value, float):
                serialized_value = AttributeValue(integer=int(value))

            serialized_attributes.append(Attribute(key=key, value=serialized_value))

        return serialized_attributes

    def _send(self, spans: List[ResourceSpan]) -> None:
        client = httpx.Client(
            headers={
                "Authorization": f"Bearer {self._options.api_key}",
                "Content-Type": "application/json",
            },
            timeout=self._options.timeout,
            follow_redirects=False,
            max_redirects=0,
        )
        response = None
        attempt = 1

        try:
            method = "POST"
            url = f"{self._options.gateway.base_url}/otlp/v1/traces"
            content = CreateTraceRequestBody(resource_spans=spans).model_dump_json()

            while attempt <= self._options.retries:
                try:
                    response = client.request(method=method, url=url, content=content)
                    response.raise_for_status()

                    break

                except Exception as exception:
                    if attempt >= self._options.retries:
                        raise exception

                    if response and response.status_code in RETRIABLE_STATUSES:
                        time.sleep(self._options.delay * (2 ** (attempt - 1)))
                    else:
                        raise exception

                finally:
                    if response:
                        response.close()

                    attempt += 1

        finally:
            client.close()



import json
from typing import Any, Dict, Optional, Sequence, Union

from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.threading import ThreadingInstrumentor
from opentelemetry.sdk import resources as otel
from opentelemetry.sdk.trace import Tracer, TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, SimpleSpanProcessor

from latitude_telemetry.env import env
from latitude_telemetry.exporter import Exporter, ExporterOptions
from latitude_telemetry.telemetry.types import (
    GatewayOptions,
    Instrumentors,
    SpanMetadata,
    SpanPrompt,
    TelemetryAttributes,
)
from latitude_telemetry.util import Model, is_package_installed, returns_like

TELEMETRY_INSTRUMENTATION_NAME = "opentelemetry.instrumentation.latitude"


_THREADING_INSTRUMENTOR = "Threading"


class InternalOptions(Model):
    gateway: Optional[GatewayOptions] = None
    retries: Optional[int] = None
    delay: Optional[float] = None
    timeout: Optional[float] = None


class TelemetryOptions(Model):
    instrumentors: Optional[Sequence[Instrumentors]] = None
    disable_batch: Optional[bool] = None
    internal: Optional[InternalOptions] = None


DEFAULT_INTERNAL_OPTIONS = InternalOptions(
    gateway=GatewayOptions(
        host=env.GATEWAY_HOSTNAME,
        port=env.GATEWAY_PORT,
        ssl=env.GATEWAY_SSL,
        api_version="v2",
    ),
    retries=3,
    delay=0.5,
    timeout=30,
)


DEFAULT_TELEMETRY_OPTIONS = TelemetryOptions(
    instrumentors=[],  # Note: Instrumentation is opt-in
    disable_batch=False,
    internal=DEFAULT_INTERNAL_OPTIONS,
)


class Telemetry:
    _options: TelemetryOptions
    _exporter: Exporter
    _tracer: TracerProvider
    _instrumentors: Dict[Union[Instrumentors, str], BaseInstrumentor]

    def __init__(self, api_key: str, options: TelemetryOptions):
        options.internal = InternalOptions(**{**dict(DEFAULT_INTERNAL_OPTIONS), **dict(options.internal or {})})
        options = TelemetryOptions(**{**dict(DEFAULT_TELEMETRY_OPTIONS), **dict(options)})

        assert options.internal is not None
        assert options.internal.gateway is not None
        assert options.internal.retries is not None
        assert options.internal.delay is not None
        assert options.internal.timeout is not None

        self._options = options
        self._exporter = Exporter(
            ExporterOptions(
                api_key=api_key,
                gateway=options.internal.gateway,
                retries=options.internal.retries,
                delay=options.internal.delay,
                timeout=options.internal.timeout,
            )
        )
        self._tracer = TracerProvider(
            resource=otel.Resource.create({otel.SERVICE_NAME: __package__ or __name__}),
        )

        if options.disable_batch:
            self._tracer.add_span_processor(SimpleSpanProcessor(self._exporter))
        else:
            self._tracer.add_span_processor(BatchSpanProcessor(self._exporter))

        self._init_instrumentors()
        self.instrument()

    def _init_instrumentors(self):
        self._instrumentors = {}

        # Threading instrumentor makes sure otel context is propagated
        self._instrumentors[_THREADING_INSTRUMENTOR] = ThreadingInstrumentor()

        if is_package_installed("aleph_alpha_client"):
            from opentelemetry.instrumentation.alephalpha import AlephAlphaInstrumentor

            self._instrumentors[Instrumentors.AlephAlpha] = AlephAlphaInstrumentor()

        if is_package_installed("anthropic"):
            from opentelemetry.instrumentation.anthropic import AnthropicInstrumentor

            self._instrumentors[Instrumentors.Anthropic] = AnthropicInstrumentor(enrich_token_usage=True)

        if is_package_installed("boto3"):
            from opentelemetry.instrumentation.bedrock import BedrockInstrumentor

            self._instrumentors[Instrumentors.Bedrock] = BedrockInstrumentor(enrich_token_usage=True)

        if is_package_installed("cohere"):
            from opentelemetry.instrumentation.cohere import CohereInstrumentor

            self._instrumentors[Instrumentors.Cohere] = CohereInstrumentor()

        if is_package_installed("google-generativeai"):
            from opentelemetry.instrumentation.google_generativeai import GoogleGenerativeAiInstrumentor

            self._instrumentors[Instrumentors.GoogleGenerativeAI] = GoogleGenerativeAiInstrumentor()

        if is_package_installed("groq"):
            from opentelemetry.instrumentation.groq import GroqInstrumentor

            self._instrumentors[Instrumentors.Groq] = GroqInstrumentor(enrich_token_usage=True)

        if is_package_installed("haystack"):
            from opentelemetry.instrumentation.haystack import HaystackInstrumentor

            self._instrumentors[Instrumentors.Haystack] = HaystackInstrumentor()

        if is_package_installed("langchain"):
            from opentelemetry.instrumentation.langchain import LangchainInstrumentor

            self._instrumentors[Instrumentors.Langchain] = LangchainInstrumentor()

        # Note: LiteLLM instrumentor does not work by itself yet, needs the provider instrumentor too
        # https://github.com/Arize-ai/openinference/issues/604
        if is_package_installed("litellm"):
            from openinference.instrumentation.litellm import LiteLLMInstrumentor

            self._instrumentors[Instrumentors.LiteLLM] = LiteLLMInstrumentor()

        if is_package_installed("llama-index") or is_package_installed("llama_index"):
            from opentelemetry.instrumentation.llamaindex import LlamaIndexInstrumentor

            self._instrumentors[Instrumentors.LlamaIndex] = LlamaIndexInstrumentor()

        if is_package_installed("mistralai"):
            from opentelemetry.instrumentation.mistralai import MistralAiInstrumentor

            self._instrumentors[Instrumentors.MistralAI] = MistralAiInstrumentor()

        if is_package_installed("ollama"):
            from opentelemetry.instrumentation.ollama import OllamaInstrumentor

            self._instrumentors[Instrumentors.Ollama] = OllamaInstrumentor()

        if is_package_installed("openai"):
            from opentelemetry.instrumentation.openai import OpenAIInstrumentor

            self._instrumentors[Instrumentors.OpenAI] = OpenAIInstrumentor(enrich_token_usage=True)

        if is_package_installed("replicate"):
            from opentelemetry.instrumentation.replicate import ReplicateInstrumentor

            self._instrumentors[Instrumentors.Replicate] = ReplicateInstrumentor()

        if is_package_installed("boto3"):
            from opentelemetry.instrumentation.sagemaker import SageMakerInstrumentor

            self._instrumentors[Instrumentors.Sagemaker] = SageMakerInstrumentor(enrich_token_usage=True)

        if is_package_installed("together"):
            from opentelemetry.instrumentation.together import TogetherAiInstrumentor

            self._instrumentors[Instrumentors.Together] = TogetherAiInstrumentor()

        if is_package_installed("transformers"):
            from opentelemetry.instrumentation.transformers import TransformersInstrumentor

            self._instrumentors[Instrumentors.Transformers] = TransformersInstrumentor()

        if is_package_installed("google-cloud-aiplatform"):
            from opentelemetry.instrumentation.vertexai import VertexAIInstrumentor

            self._instrumentors[Instrumentors.VertexAI] = VertexAIInstrumentor()

        if is_package_installed("ibm-watsonx-ai") or is_package_installed("ibm-watson-machine-learning"):
            from opentelemetry.instrumentation.watsonx import WatsonxInstrumentor

            self._instrumentors[Instrumentors.Watsonx] = WatsonxInstrumentor()

    def instrument(self, instrumentors: Optional[Sequence[Instrumentors]] = None):
        enabled_instrumentors = [_THREADING_INSTRUMENTOR, *(instrumentors or self._options.instrumentors or [])]

        for instrumentor in enabled_instrumentors:
            if (
                instrumentor in self._instrumentors
                and not self._instrumentors[instrumentor].is_instrumented_by_opentelemetry
            ):
                self._instrumentors[instrumentor].instrument(tracer_provider=self._tracer)

    def uninstrument(self):
        for instrumentor in self._instrumentors.values():
            if instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.uninstrument()

    @returns_like(Tracer.start_as_current_span)
    def span(
        self,
        name: str,
        prompt: Optional[SpanPrompt] = None,
        distinct_id: Optional[str] = None,
        metadata: Optional[SpanMetadata] = None,
    ) -> Any:
        attributes = {}

        if prompt:
            attributes[TelemetryAttributes.Prompt] = prompt.model_dump_json()

        if distinct_id:
            attributes[TelemetryAttributes.DistinctID] = distinct_id

        if metadata:
            attributes[TelemetryAttributes.Metadata] = json.dumps(metadata)

        tracer = self._tracer.get_tracer(TELEMETRY_INSTRUMENTATION_NAME)

        return tracer.start_as_current_span(name, attributes=attributes)  # type: ignore



from typing import List, cast
from unittest import mock

import httpx

from latitude_telemetry import SpanPrompt
from tests.utils import TestCase


class TestSpan(TestCase):
    def test_success(self):
        endpoint = "/otlp/v1/traces"
        endpoint_mock = self.gateway_mock.post(endpoint).mock()

        with self.telemetry.span(
            name="first",
            prompt=SpanPrompt(
                uuid="prompt-uuid",
                version_uuid="version-uuid",
                parameters={"parameter": "value"},
            ),
            distinct_id="distinct-id",
            metadata={"key": "value"},
        ):
            with self.telemetry.span("second"):
                pass

        requests = cast(List[httpx.Request], [request for request, _ in endpoint_mock.calls])  # type: ignore

        self.assert_requested(
            requests[0],
            method="POST",
            endpoint=endpoint,
            body=self.create_instrumentation_request(
                name="second",
                parentSpanId=mock.ANY,
                attributes=[],
            ),
        )
        self.assert_requested(
            requests[1],
            method="POST",
            endpoint=endpoint,
            body=self.create_instrumentation_request(
                name="first",
                attributes=[
                    {
                        "key": "latitude.prompt",
                        "value": {
                            "stringValue": '{"uuid":"prompt-uuid","versionUuid":"version-uuid","parameters":{"parameter":"value"}}'  # noqa: E501
                        },
                    },
                    {"key": "latitude.distinctId", "value": {"stringValue": "distinct-id"}},
                    {"key": "latitude.metadata", "value": {"stringValue": '{"key": "value"}'}},
                ],
            ),
        )
        self.assertEqual(endpoint_mock.call_count, 2)



from typing import Any, Dict, Optional

from latitude_telemetry.util import Field, Model, StrEnum


class Instrumentors(StrEnum):
    AlephAlpha = "alephalpha"
    Anthropic = "anthropic"
    Bedrock = "bedrock"
    Cohere = "cohere"
    GoogleGenerativeAI = "googlegenerativeai"
    Groq = "groq"
    Haystack = "haystack"
    Langchain = "langchain"
    LiteLLM = "litellm"
    LlamaIndex = "llamaindex"
    MistralAI = "mistralai"
    Ollama = "ollama"
    OpenAI = "openai"
    Replicate = "replicate"
    Sagemaker = "sagemaker"
    Together = "together"
    Transformers = "transformers"
    VertexAI = "vertexai"
    Watsonx = "watsonx"


class TelemetryAttributes(StrEnum):
    Root = "latitude"
    Span = f"{Root}.span"
    Prompt = f"{Root}.prompt"
    DistinctID = f"{Root}.distinctId"
    Metadata = f"{Root}.metadata"


class SpanPrompt(Model):
    uuid: str
    version_uuid: Optional[str] = Field(default=None, alias=str("versionUuid"))
    parameters: Optional[Dict[str, Any]] = None


SpanMetadata = Dict[str, Any]


class GatewayOptions(Model):
    host: str
    port: int
    ssl: bool
    api_version: str

    @property
    def protocol(self) -> str:
        return "https" if self.ssl else "http"

    @property
    def base_url(self) -> str:
        return f"{self.protocol}://{self.host}:{self.port}/api/{self.api_version}"



import json
import unittest
from typing import Any, Dict, Optional
from unittest import mock

import httpx
import respx

from latitude_telemetry import GatewayOptions, InternalOptions, Telemetry, TelemetryOptions


class TestCase(unittest.TestCase):
    telemetry: Telemetry

    def setUp(self):
        self.maxDiff = None

        self.internal_options = InternalOptions(
            gateway=GatewayOptions(
                host="fake-host.com",
                port=443,
                ssl=True,
                api_version="v2",
            ),
            retries=3,
            delay=0,
            timeout=0.5,
        )
        self.api_key = "fake-api-key"
        self.base_url = "https://fake-host.com/api/v2"

        self.gateway_mock = respx.MockRouter(
            assert_all_called=False,
            assert_all_mocked=True,
            base_url=self.base_url,
        )
        self.gateway_mock.start()
        self.gateway_mock.reset()
        self.gateway_mock.clear()

        self.telemetry = Telemetry(
            self.api_key,
            TelemetryOptions(
                instrumentors=[],
                disable_batch=True,
                internal=self.internal_options,
            ),
        )

    def tearDown(self):
        self.telemetry.uninstrument()
        self.gateway_mock.stop()

    def assert_requested(
        self,
        request: httpx.Request,
        method: str,
        endpoint: str,
        headers: Optional[Dict[str, Any]] = None,
        body: Optional[Dict[str, Any]] = None,
    ):
        self.assertEqual(request.method, method)
        self.assertEqual(request.url, f"{self.base_url}{endpoint}")
        self.assertDictContainsSubset(
            {**{"authorization": f"Bearer {self.api_key}"}, **(headers or {})},
            dict(request.headers),
        )
        try:
            self.assertEqual(
                json.loads(request.content),
                {**(body or {})},
            )
        except json.JSONDecodeError:
            self.assertEqual(None, body)

    def create_instrumentation_request(self, **kwargs: Any) -> Dict[str, Any]:
        return {
            "resourceSpans": [
                {
                    "resource": {
                        "attributes": [
                            {"key": "telemetry.sdk.language", "value": {"stringValue": "python"}},
                            {"key": "telemetry.sdk.name", "value": {"stringValue": "opentelemetry"}},
                            {"key": "telemetry.sdk.version", "value": {"stringValue": "1.29.0"}},
                            {"key": "service.name", "value": {"stringValue": "latitude_telemetry.telemetry"}},
                        ]
                    },
                    "scopeSpans": [
                        {
                            "spans": [
                                {
                                    "traceId": mock.ANY,
                                    "spanId": mock.ANY,
                                    "name": "",
                                    "kind": mock.ANY,
                                    "startTimeUnixNano": mock.ANY,
                                    "endTimeUnixNano": mock.ANY,
                                    "status": {"code": mock.ANY},
                                    "events": [],
                                    "links": [],
                                    "attributes": [],
                                    **kwargs,
                                }
                            ]
                        }
                    ],
                }
            ]
        }



from typing import List, Optional

from latitude_telemetry.util import Field, Model


class AttributeValue(Model):
    string: Optional[str] = Field(default=None, alias=str("stringValue"))
    integer: Optional[int] = Field(default=None, alias=str("intValue"))
    boolean: Optional[bool] = Field(default=None, alias=str("boolValue"))


class Attribute(Model):
    key: str
    value: AttributeValue


class Resource(Model):
    attributes: List[Attribute]


class Status(Model):
    code: int
    message: Optional[str] = None


class Event(Model):
    name: str
    time: str = Field(alias=str("timeUnixNano"))
    attributes: Optional[List[Attribute]] = None


class Link(Model):
    trace_id: str = Field(alias=str("traceId"))
    span_id: str = Field(alias=str("spanId"))
    attributes: Optional[List[Attribute]] = None


class Span(Model):
    trace_id: str = Field(alias=str("traceId"))
    span_id: str = Field(alias=str("spanId"))
    parent_span_id: Optional[str] = Field(default=None, alias=str("parentSpanId"))
    name: str
    kind: int
    start_time: str = Field(alias=str("startTimeUnixNano"))
    end_time: Optional[str] = Field(default=None, alias=str("endTimeUnixNano"))
    status: Optional[Status] = None
    events: Optional[List[Event]] = None
    links: Optional[List[Link]] = None
    attributes: Optional[List[Attribute]] = None


class ScopeSpan(Model):
    spans: List[Span]


class ResourceSpan(Model):
    resource: Resource
    scope_spans: List[ScopeSpan] = Field(alias=str("scopeSpans"))


class CreateTraceRequestBody(Model):
    resource_spans: List[ResourceSpan] = Field(alias=str("resourceSpans"))
